import { ComputeBudgetProgram, Transaction, TransactionMessage, VersionedTransaction, } from '@solana/web3.js';
import { chunks } from './accountFetchCache';
import bs58 from 'bs58';
import { ProgramError } from './anchorError';
import { consoleLogger } from './logger';
export class Tx {
    instrs = [];
    _feePayer;
    _startPrio = 'NONE';
    _maxPrio = 'VeryHigh';
    _commitment = 'confirmed';
    _addressLookupTableAccounts = [];
    _sendOptions = { skipPreflight: true };
    _timeoutMillis = HIGH_END_TIMEOUT_MILLIS;
    _log = consoleLogger;
    _doNotRetryAt = new Map();
    _prioFeeEstimateOpts = {
        includeVote: false,
        recommended: true,
    };
    _signers = [];
    constructor(p) {
        if (p.ixs) {
            this.instrs = p.ixs;
        }
        this._feePayer = p.feePayer;
        if (p.log) {
            this._log = p.log;
        }
    }
    add(...ixs) {
        this.instrs.push(...ixs);
        return this;
    }
    startPrio(p) {
        this._prioFeeEstimateOpts.recommended = false;
        this._startPrio = p;
        return this;
    }
    maxPrio(p) {
        this._maxPrio = p;
        return this;
    }
    commitment(c) {
        this._commitment = c;
        return this;
    }
    lookupTable(as) {
        this._addressLookupTableAccounts = as;
        return this;
    }
    sendOptions(o) {
        this._sendOptions = o;
        return this;
    }
    timeoutMillis(m) {
        this._timeoutMillis = m;
        return this;
    }
    log(l) {
        this._log = l;
        return this;
    }
    doNotRetryAt(errCodeToDesc) {
        this._doNotRetryAt = errCodeToDesc;
        return this;
    }
    prioFeeEstOpts(opts) {
        let opts1 = opts;
        if (this._startPrio != 'NONE') {
            let { recommended, ...rest } = opts;
            opts1 = rest;
        }
        this._prioFeeEstimateOpts = opts1;
        return this;
    }
    signers(...signers) {
        this._signers.push(...signers);
        return this;
    }
    async send(connection) {
        const consTx = async (instrs) => {
            const signers = this._signers;
            //if (!(signers.map(s => s.publicKey.toString()).includes(this._feePayer.toString())))
            const blockhash = (await connection.getLatestBlockhash('confirmed'))
                .blockhash;
            // feePayer ?? signers[0].publicKey
            const payerKey = this._feePayer;
            if (this._addressLookupTableAccounts.length > 0) {
                const messageV0 = new TransactionMessage({
                    payerKey,
                    recentBlockhash: blockhash,
                    instructions: [...instrs],
                }).compileToV0Message(this._addressLookupTableAccounts);
                const tx = new VersionedTransaction(messageV0);
                tx.sign(signers);
                return Buffer.from(tx.serialize());
            }
            const tx = new Transaction({
                feePayer: payerKey,
                recentBlockhash: blockhash,
            }).add(...instrs);
            tx.sign(...signers);
            return tx.serialize();
        };
        const recur = async (priorityLevel) => {
            let instrs = this.instrs;
            let txSer = await consTx(instrs);
            if (priorityLevel !== 'NONE') {
                const { priorityFeeEstimate } = await getPriorityFeeEstimate(connection, txSer, priorityLevel, this._log, this._prioFeeEstimateOpts);
                if (priorityFeeEstimate) {
                    const computePriceIx = ComputeBudgetProgram.setComputeUnitPrice({
                        // @ts-ignore
                        microLamports: Math.trunc(priorityFeeEstimate),
                    });
                    txSer = await consTx([...instrs, computePriceIx]);
                }
            }
            const log = this._log;
            try {
                return await sendAndConfirmWithRetry(connection, txSer, this._sendOptions, this._commitment, this._timeoutMillis, log);
            }
            catch (e) {
                // @ts-ignore
                let baseMsg = `failed at prio ${priorityLevel}: ${e} ${e.stack}`;
                if (priorityLevel == this._maxPrio) {
                    log.info(`${baseMsg} hard failed`);
                    throw e;
                }
                const wrappedE = ProgramError.parse(e, this._doNotRetryAt);
                if (wrappedE) {
                    log.info(`got do not retry err: ${wrappedE}. aborting`);
                    throw e;
                }
                let newPrio;
                if (priorityLevel == 'Low') {
                    newPrio = 'Medium';
                }
                else if (priorityLevel == 'Medium') {
                    newPrio = 'High';
                }
                else if (priorityLevel == 'High') {
                    newPrio = 'VeryHigh';
                }
                else {
                    newPrio = 'Low';
                }
                log.info(`${baseMsg}. retrying at ${newPrio}`);
                return await recur(newPrio);
            }
        };
        return await recur(this._startPrio);
    }
}
async function sleep(ms) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}
async function promiseAllInOrder(it) {
    let ret = [];
    for (const i of it) {
        ret.push(await i());
    }
    return ret;
}
export async function sendInstructions(provider, instructions, signers = [], payer = provider.wallet.publicKey, commitment = 'confirmed', idlErrors = new Map(), log = consoleLogger) {
    if (instructions.length == 0) {
        return '';
    }
    let tx = new Transaction();
    tx.recentBlockhash = (await provider.connection.getLatestBlockhash()).blockhash;
    tx.feePayer = payer || provider.wallet.publicKey;
    tx.add(...instructions);
    if (signers.length > 0) {
        tx.partialSign(...signers);
    }
    if (tx.feePayer.equals(provider.wallet.publicKey) ||
        tx.instructions.some((ix) => ix.keys.some((key) => key.isSigner && key.pubkey.equals(provider.wallet.publicKey)))) {
        tx = await provider.wallet.signTransaction(tx);
    }
    try {
        const { txid } = await sendAndConfirmWithRetry(provider.connection, tx.serialize(), {
            skipPreflight: true,
        }, commitment);
        return txid;
    }
    catch (e) {
        log.error(e);
        if (e.logs) {
            log.error(`Logs: ${e.logs}`);
        }
        const wrappedE = ProgramError.parse(e, idlErrors);
        throw wrappedE == null ? e : wrappedE;
    }
}
function truthy(value) {
    return !!value;
}
export async function sendMultipleInstructions(provider, instructionGroups, signerGroups, payer, finality = 'confirmed', idlErrors = new Map(), log = consoleLogger) {
    const recentBlockhash = (await provider.connection.getLatestBlockhash('confirmed')).blockhash;
    const ixAndSigners = instructionGroups
        .map((instructions, i) => {
        const signers = signerGroups[i];
        return {
            instructions,
            signers,
        };
    })
        .filter(({ instructions }) => instructions.length > 0);
    const txns = ixAndSigners.map(({ instructions }) => {
        const tx = new Transaction({
            feePayer: payer || provider.wallet.publicKey,
            recentBlockhash,
        });
        tx.add(...instructions);
        return tx;
    });
    const txnsSignedByWallet = await provider.wallet.signAllTransactions(txns);
    const txnsSigned = txnsSignedByWallet
        .map((tx, index) => {
        const signers = ixAndSigners[index].signers;
        if (signers.length > 0) {
            tx.partialSign(...signers);
        }
        return tx;
    })
        .map((tx) => tx.serialize());
    log.info('Sending multiple transactions...');
    try {
        return await promiseAllInOrder(txnsSigned.map((txn) => async () => {
            const { txid } = await sendAndConfirmWithRetry(provider.connection, txn, {
                skipPreflight: true,
            }, finality);
            return txid;
        }));
    }
    catch (e) {
        log.error(e);
        const wrappedE = ProgramError.parse(e, idlErrors);
        throw wrappedE == null ? e : wrappedE;
    }
}
export async function execute(program, provider, command, payer = provider.wallet.publicKey, commitment) {
    const { instructions, signers, output } = command;
    const errors = program.idl.errors?.reduce((acc, err) => {
        acc.set(err.code, `${err.name}: ${err.msg}`);
        return acc;
    }, new Map());
    if (instructions.length > 0) {
        const txid = await sendInstructions(provider, instructions, signers, payer, commitment, errors);
        return { txid, ...output };
    }
    // @ts-ignore
    return output;
}
export async function executeBig(program, provider, command, payer = provider.wallet.publicKey, finality) {
    const { instructions, signers, output } = command;
    const errors = program.idl.errors?.reduce((acc, err) => {
        acc.set(err.code, `${err.name}: ${err.msg}`);
        return acc;
    }, new Map());
    if (instructions.length > 0) {
        const txids = await sendMultipleInstructions(provider, instructions, signers, payer || provider.wallet.publicKey, finality, errors);
        return {
            ...output,
            txids: Array.from(txids),
        };
    }
    // @ts-ignore
    return output;
}
function getUnixTimeSec() {
    return new Date().valueOf() / 1000;
}
export const getPriorityFeeEstimate = async (connection, transaction, priorityLevel = 'NONE', log = consoleLogger, options = { includeVote: true, recommended: true }) => {
    let opts = options;
    if (!options.recommended) {
        opts = {
            includeVote: options.includeVote,
            priorityLevel: priorityLevel,
        };
    }
    const response = await fetch(connection.rpcEndpoint, {
        method: 'POST',
        headers: { 'Content-Type': 'application/json' },
        body: JSON.stringify({
            jsonrpc: '2.0',
            id: '1',
            method: 'getPriorityFeeEstimate',
            params: [
                {
                    transaction: bs58.encode(transaction),
                    options: opts,
                },
            ],
        }),
    });
    if (response.status != 200) {
        const err = `getPriorityFeeEstimate failed ${await response.text()}`;
        log.error(err);
        throw new Error(err);
    }
    const data = await response.json();
    let priorityFeeEstimate = 50000;
    if (!data.result || data.result.priorityFeeEstimate == undefined) {
        log.warn(`getPriorityFeeEstimate unknown payload ${JSON.stringify(data)}`);
    }
    else {
        priorityFeeEstimate = data.result.priorityFeeEstimate;
    }
    // @ts-ignore
    log.info(`Fee in function for ${priorityLevel}: ${priorityFeeEstimate}`);
    // @ts-ignore
    return { priorityFeeEstimate };
};
export const sendAndConfirmWithRetryAndPrioFees = async (connection, instrs, signers, sendOptions, commitment, feePayer, priorityLevel = 'NONE', timeoutMillis = HIGH_END_TIMEOUT_MILLIS, log = consoleLogger, addressLookupTableAccounts = [], doNotRetryAt = new Map(), maxPriorityLevel = 'VeryHigh') => {
    const consTx = async (instrs) => {
        const blockhash = (await connection.getLatestBlockhash('confirmed'))
            .blockhash;
        const payerKey = feePayer ?? signers[0].publicKey;
        if (addressLookupTableAccounts.length > 0) {
            const messageV0 = new TransactionMessage({
                payerKey,
                recentBlockhash: blockhash,
                instructions: [...instrs],
            }).compileToV0Message(addressLookupTableAccounts);
            const tx = new VersionedTransaction(messageV0);
            tx.sign(signers);
            return Buffer.from(tx.serialize());
        }
        const tx = new Transaction({
            feePayer: payerKey,
            recentBlockhash: blockhash,
        }).add(...instrs);
        tx.sign(...signers);
        return tx.serialize();
    };
    let txSer = await consTx(instrs);
    if (priorityLevel !== 'NONE') {
        const feeEstimate = await getPriorityFeeEstimate(connection, txSer, priorityLevel, log);
        const computePriceIx = ComputeBudgetProgram.setComputeUnitPrice({
            // @ts-ignore
            microLamports: Math.trunc(feeEstimate.priorityFeeEstimate),
        });
        txSer = await consTx([...instrs, computePriceIx]);
    }
    try {
        return await sendAndConfirmWithRetry(connection, txSer, sendOptions, commitment, timeoutMillis, log);
    }
    catch (e) {
        // @ts-ignore
        let baseMsg = `failed at prio ${priorityLevel}: ${e} ${e.stack}`;
        if (priorityLevel == maxPriorityLevel) {
            log.info(`${baseMsg} hard failed`);
            throw e;
        }
        const wrappedE = ProgramError.parse(e, doNotRetryAt);
        if (wrappedE) {
            log.info(`got do not retry err: ${wrappedE}. aborting`);
            throw e;
        }
        let newPrio;
        if (priorityLevel == 'Low') {
            newPrio = 'Medium';
        }
        else if (priorityLevel == 'Medium') {
            newPrio = 'High';
        }
        else if (priorityLevel == 'High') {
            newPrio = 'VeryHigh';
        }
        else {
            newPrio = 'Low';
        }
        log.info(`${baseMsg}. retrying at ${newPrio}`);
        return await sendAndConfirmWithRetryAndPrioFees(connection, instrs, signers, sendOptions, commitment, feePayer, newPrio, timeoutMillis, log, addressLookupTableAccounts);
    }
};
export const sendMultipleInstructionsWithPrioFee = async (provider, instructionGroups, signerGroups, payer, priorityLevel = 'NONE', finality = 'confirmed', timeout = HIGH_END_TIMEOUT_MILLIS, idlErrors = new Map(), log = consoleLogger) => {
    const txSigs = instructionGroups.map((_) => '');
    let currI = 0;
    const origPriorityLevel = priorityLevel;
    const sendMultipleTxs = async (instructionGroups, signerGroups, priorityLevel) => {
        const recentBlockhash = (await provider.connection.getLatestBlockhash('confirmed')).blockhash;
        const ixAndSigners = instructionGroups
            .map((instructions, i) => {
            const signers = signerGroups[i];
            return {
                instructions,
                signers,
            };
        })
            .filter(({ instructions }) => instructions.length > 0);
        const txns = ixAndSigners.map(({ instructions }) => {
            const tx = new Transaction({
                feePayer: payer || provider.wallet.publicKey,
                recentBlockhash,
            });
            tx.add(...instructions);
            return tx;
        });
        const txnsSignedByWallet = await provider.wallet.signAllTransactions(txns);
        let txnsSigned = txnsSignedByWallet.map((tx, index) => {
            const signers = ixAndSigners[index].signers;
            if (signers.length > 0) {
                tx.partialSign(...signers);
            }
            return tx;
        });
        if (priorityLevel !== 'NONE') {
            const feeEstimates = await Promise.all(txnsSigned.map((tx, i) => getPriorityFeeEstimate(provider.connection, tx.serialize(), i == currI ? priorityLevel : origPriorityLevel)));
            for (let i = 0; i < feeEstimates.length; i += 1) {
                const feeEstimate = feeEstimates[i];
                const computePriceIx = ComputeBudgetProgram.setComputeUnitPrice({
                    // @ts-ignore
                    microLamports: Math.trunc(feeEstimate.priorityFeeEstimate),
                });
                const tx = txnsSigned[i];
                tx.add(computePriceIx);
                // resign
                const signers = ixAndSigners[i].signers;
                if (signers.length > 0) {
                    tx.partialSign(...signers);
                }
            }
        }
        log.debug('Sending multiple transactions...');
        try {
            return await promiseAllInOrder(txnsSigned.map((txn, i) => async () => {
                try {
                    const { txid } = await sendAndConfirmWithRetry(provider.connection, txn.serialize(), {
                        skipPreflight: true,
                    }, finality, timeout);
                    txSigs[currI] = txid;
                    currI += 1;
                    return txid;
                }
                catch (e) {
                    log.error(`txidx ${currI} err ${e}`);
                    throw e;
                }
            }));
        }
        catch (e) {
            log.error(e);
            const wrappedE = ProgramError.parse(e, idlErrors);
            throw wrappedE == null ? e : wrappedE;
        }
    };
    const recur = async (instructionGroups, signerGroups, priorityLevel) => {
        try {
            await sendMultipleTxs(instructionGroups, signerGroups, priorityLevel);
            return;
        }
        catch (e) {
            if (e instanceof ProgramError) {
                log.error(`aborting. got program error ${e}`);
                throw e;
            }
            // @ts-ignore
            let baseMsg = `failed at prio ${priorityLevel}: ${e} ${e.stack}`;
            let newPrio;
            if (priorityLevel === 'NONE') {
                newPrio = 'Low';
            }
            else if (priorityLevel == 'Low') {
                newPrio = 'Medium';
            }
            else if (priorityLevel == 'Medium') {
                newPrio = 'High';
            }
            else if (priorityLevel == 'High') {
                newPrio = 'VeryHigh';
            }
            const msg = `${baseMsg}. ${newPrio ? `retrying at ${newPrio}` : `hard failed`}`;
            log.info(msg);
            if (!newPrio) {
                throw e;
            }
            await recur(instructionGroups.slice(currI), signerGroups.slice(currI), newPrio);
            return;
        }
    };
    await recur(instructionGroups, signerGroups, priorityLevel);
    return txSigs;
};
const SEND_TRANSACTION_INTERVAL = 10;
export const awaitTransactionSignatureConfirmation = async (txid, timeoutMillis, connection, commitment = 'recent', queryStatus = false, log = consoleLogger) => {
    let done = false;
    let status = {
        slot: 0,
        confirmations: 0,
        err: null,
    };
    let subId = 0;
    status = await new Promise(async (resolve, reject) => {
        let t;
        function setDone() {
            done = true;
            clearTimeout(t);
        }
        t = setTimeout(() => {
            if (done) {
                return;
            }
            setDone();
            log.error('Rejecting for timeout...');
            reject({ timeout: true });
        }, timeoutMillis);
        try {
            subId = connection.onSignature(txid, (result, context) => {
                status = {
                    err: result.err,
                    slot: context.slot,
                    confirmations: 0,
                };
                setDone();
                if (result.err) {
                    log.error('Rejected via websocket', result.err);
                    reject(status);
                }
                else {
                    resolve(status);
                }
            }, commitment);
        }
        catch (e) {
            log.error('WS error in setup', txid, e);
            if (!queryStatus) {
                reject(e);
            }
        }
        while (!done && queryStatus) {
            ;
            (async () => {
                try {
                    const signatureStatuses = await connection.getSignatureStatuses([
                        txid,
                    ]);
                    status = signatureStatuses && signatureStatuses.value[0];
                    if (!done) {
                        if (!status) {
                        }
                        else if (status.err) {
                            log.error('REST error for', txid, status);
                            setDone();
                            reject(status.err);
                        }
                        else if (!status.confirmations && !status.confirmationStatus) {
                            log.warn('REST no confirmations for', txid, status);
                        }
                        else {
                            log.info('REST confirmation for', txid, status);
                            if (!status.confirmationStatus ||
                                status.confirmationStatus == commitment) {
                                setDone();
                                resolve(status);
                            }
                        }
                    }
                }
                catch (e) {
                    if (!done) {
                        log.error('REST connection error: txid', txid, e);
                    }
                }
            })();
            await sleep(2000);
        }
    });
    if (
    //@ts-ignore
    connection._signatureSubscriptions &&
        //@ts-ignore
        connection._signatureSubscriptions[subId]) {
        connection.removeSignatureListener(subId);
    }
    done = true;
    return status;
};
async function simulateTransaction(connection, transaction, commitment) {
    // @ts-ignore
    transaction.recentBlockhash = await connection._recentBlockhash(
    // @ts-ignore
    connection._disableBlockhashCaching);
    const signData = transaction.serializeMessage();
    // @ts-ignore
    const wireTransaction = transaction._serialize(signData);
    const encodedTransaction = wireTransaction.toString('base64');
    const config = { encoding: 'base64', commitment };
    const args = [encodedTransaction, config];
    // @ts-ignore
    const res = await connection._rpcRequest('simulateTransaction', args);
    if (res.error) {
        throw new Error('failed to simulate transaction: ' + res.error.message);
    }
    return res.result;
}
// DO NOT CHANGE THE NUMBERS BELOW:
// if you have retry logic and its set to retry after some N sec wait time << the lowest range of confirmation times
// (400-600ms/slot expiring after 151 slots -> 400 * 151 = 60.4s. so lets say you set it to 30s)
// you could wind up double mining your tx because you assumed the previous retry iteration timed out when in reality
// the tx was actually mined
const HIGH_END_TIMEOUT_MILLIS = 91 * 1000; // 91s -> 600ms high end slot time * 151 slots = 90.4s
const MAX_TIMEOUT_MILLIS = 120 * 1000; // 120s -> not sure where this num comes from
/*
    A validator has up to 120s to accept the transaction and send it into a block.
    If it doesn’t happen within that timeframe, your transaction is dropped and you’ll need
    to send the transaction again. You can get the transaction signature and periodically
    Ping the network for that transaction signature. If you never get anything back,
    that means it’s definitely been dropped. If you do get a response back, you can keep pinging
    that means it’s definitely been dropped. If you do get a response back, you can keep pinging
    until it’s gone to a confirmed status to move on.
  */
export async function sendAndConfirmWithRetry(connection, txn, sendOptions, commitment, timeoutMillis = HIGH_END_TIMEOUT_MILLIS, log = consoleLogger, maxSupportedTransactionVersion) {
    let done = false;
    let slot = 0;
    const txid = await connection.sendRawTransaction(txn, sendOptions);
    const startTime = Date.now();
    (async () => {
        while (!done && Date.now() - startTime < timeoutMillis) {
            log.debug(`sending raw tx now - startTime ${Date.now() - startTime} < ${timeoutMillis}`);
            await connection.sendRawTransaction(txn, sendOptions);
            await sleep(500);
        }
    })();
    try {
        const confirmation = await awaitTransactionSignatureConfirmation(txid, timeoutMillis, connection, commitment, true);
        if (!confirmation)
            throw new Error('Timed out awaiting confirmation on transaction');
        if (confirmation.err) {
            const tx = await connection.getTransaction(txid, {
                commitment: 'confirmed',
                maxSupportedTransactionVersion,
            });
            log.error(tx?.meta?.logMessages?.join('\n'));
            log.error(confirmation.err);
            throw new Error('Transaction failed: Custom instruction error');
        }
        slot = confirmation?.slot || 0;
    }
    catch (err) {
        log.error('Error caught', err);
        if (err.timeout) {
            throw new Error('Timed out awaiting confirmation on transaction');
        }
        const tx = await connection.getTransaction(txid, {
            commitment: 'confirmed',
            maxSupportedTransactionVersion,
        });
        if (tx && tx.meta && tx.meta.logMessages) {
            log.error(tx.meta.logMessages.join('\n'));
        }
        if (err.err) {
            throw err.err;
        }
        throw err;
    }
    finally {
        done = true;
    }
    log.info('Latency', txid, Date.now() - startTime);
    return { txid };
}
export function stringToTransaction(solanaTransaction) {
    return Transaction.from(Buffer.from(solanaTransaction));
}
export function bufferToTransaction(solanaTransaction) {
    return Transaction.from(solanaTransaction);
}
async function withRetries(tries, input, log = consoleLogger) {
    for (let i = 0; i < tries; i++) {
        try {
            return await input();
        }
        catch (e) {
            log.debug(`Retrying ${i}...`, e);
        }
    }
    throw new Error('Failed after retries');
}
const TX_BATCH_SIZE = 200;
export async function bulkSendTransactions(provider, txs, onProgress, triesRemaining = 10) {
    let ret = [];
    // attempt to chunk by blockhash bounds (so signing doesn't take too long)
    for (let chunk of chunks(txs, TX_BATCH_SIZE)) {
        const thisRet = [];
        // Continually send in bulk while resetting blockhash until we send them all
        while (true) {
            const recentBlockhash = await withRetries(5, () => provider.connection.getLatestBlockhash('confirmed'));
            const signedTxs = await Promise.all(chunk.map(async (tx) => {
                tx.recentBlockhash = recentBlockhash.blockhash;
                // @ts-ignore
                const signed = await provider.wallet.signTransaction(tx);
                tx.signatures[0].signature;
                return signed;
            }));
            const txsWithSigs = signedTxs.map((tx, index) => {
                return {
                    transaction: chunk[index],
                    sig: bs58.encode(tx.signatures[0].signature),
                };
            });
            const confirmedTxs = await bulkSendRawTransactions(provider.connection, signedTxs.map((s) => s.serialize()), ({ totalProgress, ...rest }) => onProgress &&
                onProgress({
                    ...rest,
                    totalProgress: totalProgress + ret.length + thisRet.length,
                }), recentBlockhash.lastValidBlockHeight, 
            // Hail mary, try with preflight enabled. Sometimes this causes
            // errors that wouldn't otherwise happen
            triesRemaining != 1);
            thisRet.push(...confirmedTxs);
            if (confirmedTxs.length == signedTxs.length) {
                break;
            }
            const retSet = new Set(thisRet);
            chunk = txsWithSigs
                .filter(({ sig }) => !retSet.has(sig))
                .map(({ transaction }) => transaction);
            triesRemaining--;
            if (triesRemaining <= 0) {
                throw new Error(`Failed to submit all txs after blockhashes expired, ${signedTxs.length - confirmedTxs.length} remain`);
            }
        }
        ret.push(...thisRet);
    }
    return ret;
}
// Returns the list of succesfully sent txns
// NOTE: The return signatures are ordered by confirmation, not by order they are passed
// This list should be in order. Seom txns may fail
// due to blockhash exp
export async function bulkSendRawTransactions(connection, txs, onProgress, lastValidBlockHeight, skipPreflight = true, log = consoleLogger) {
    const txBatchSize = TX_BATCH_SIZE;
    let totalProgress = 0;
    const ret = [];
    if (!lastValidBlockHeight) {
        const blockhash = await withRetries(5, () => connection.getLatestBlockhash('confirmed'));
        lastValidBlockHeight = blockhash.lastValidBlockHeight;
    }
    for (let chunk of chunks(txs, txBatchSize)) {
        let currentBatchProgress = 0;
        let pendingCount = chunk.length;
        let txids = [];
        let lastRetry = 0;
        while (pendingCount > 0) {
            if ((await withRetries(5, () => connection.getBlockHeight())) >
                lastValidBlockHeight) {
                return ret;
            }
            // only resend txs every 4s
            if (lastRetry < new Date().valueOf() - 4 * 1000) {
                lastRetry = new Date().valueOf();
                txids = [];
                for (const tx of chunk) {
                    const txid = await connection.sendRawTransaction(tx, {
                        skipPreflight,
                    });
                    txids.push(txid);
                }
            }
            const statuses = await getAllTxns(connection, txids);
            const completed = statuses.filter((status) => status !== null);
            totalProgress += completed.length;
            currentBatchProgress += completed.length;
            onProgress &&
                onProgress({
                    totalProgress: totalProgress,
                    currentBatchProgress: currentBatchProgress,
                    currentBatchSize: txBatchSize,
                });
            const failures = completed
                .map((status) => status !== null && status.meta?.err)
                .filter(truthy);
            if (failures.length > 0) {
                log.error(failures);
                throw new Error('Failed to run txs');
            }
            ret.push(...txids
                .map((txid, idx) => (statuses[idx] == null ? null : txid))
                .filter(truthy));
            chunk = chunk.filter((_, index) => statuses[index] === null);
            txids = txids.filter((_, index) => statuses[index] === null);
            pendingCount -= completed.length;
            await sleep(1000); // Wait one seconds before querying again
        }
    }
    return ret;
}
const MAX_GET_SIGNATURE_STATUSES_QUERY_ITEMS = 200;
async function getAllTxns(connection, txids) {
    return (await Promise.all(chunks(txids, MAX_GET_SIGNATURE_STATUSES_QUERY_ITEMS).map((txids) => connection.getTransactions(txids, {
        maxSupportedTransactionVersion: 0,
        commitment: 'confirmed',
    })))).flat();
}
class SendIxError extends Error {
    logs;
    constructor(originalError) {
        //The newlines don't actually show up correctly in chai's assertion error, but at least
        // we have all the information and can just replace '\n' with a newline manually to see
        // what's happening without having to change the code.
        const logs = originalError.logs?.join('\n') || 'error had no logs';
        super(originalError.message + '\nlogs:\n' + logs);
        this.stack = originalError.stack;
        this.logs = logs;
    }
}
//# sourceMappingURL=transaction.js.map