/// <reference types="node" />
import { AnchorProvider, Program, Provider } from '@coral-xyz/anchor';
import { AddressLookupTableAccount, Commitment, Connection, Finality, PublicKey, SendOptions, SignatureStatus, Signer, Transaction, TransactionInstruction, TransactionSignature } from '@solana/web3.js';
import { SplUtilsLogger } from './logger';
export declare class Tx {
    instrs: TransactionInstruction[];
    _feePayer: PublicKey;
    _startPrio: PriorityLevel;
    _maxPrio: PriorityLevel;
    _commitment: Commitment;
    _addressLookupTableAccounts: AddressLookupTableAccount[];
    _sendOptions: SendOptions;
    _timeoutMillis: number;
    _log: SplUtilsLogger;
    _doNotRetryAt: Map<number, string>;
    _prioFeeEstimateOpts: PrioFeeEstimateOpts;
    _signers: Signer[];
    constructor(p: {
        feePayer: PublicKey;
        ixs?: TransactionInstruction[];
        log?: SplUtilsLogger;
    });
    add(...ixs: TransactionInstruction[]): Tx;
    startPrio(p: PriorityLevel): Tx;
    maxPrio(p: PriorityLevel): Tx;
    commitment(c: Commitment): Tx;
    lookupTable(as: AddressLookupTableAccount[]): Tx;
    sendOptions(o: SendOptions): Tx;
    timeoutMillis(m: number): Tx;
    log(l: SplUtilsLogger): Tx;
    doNotRetryAt(errCodeToDesc: Map<number, string>): Tx;
    prioFeeEstOpts(opts: PrioFeeEstimateOpts): Tx;
    signers(...signers: Signer[]): Tx;
    send(connection: Connection): Promise<{
        txid: string;
    }>;
}
export interface InstructionResult<A> {
    instructions: TransactionInstruction[];
    signers: Signer[];
    output: A;
}
export interface BigInstructionResult<A> {
    instructions: TransactionInstruction[][];
    signers: Signer[][];
    output: A;
}
export declare function sendInstructions(provider: AnchorProvider, instructions: TransactionInstruction[], signers?: Signer[], payer?: PublicKey, commitment?: Commitment, idlErrors?: Map<number, string>, log?: SplUtilsLogger): Promise<string>;
export declare function sendMultipleInstructions(provider: AnchorProvider, instructionGroups: TransactionInstruction[][], signerGroups: Signer[][], payer?: PublicKey, finality?: Finality, idlErrors?: Map<number, string>, log?: SplUtilsLogger): Promise<Iterable<string>>;
export declare function execute<Output>(program: Program<any>, provider: AnchorProvider, command: InstructionResult<Output>, payer?: PublicKey, commitment?: Commitment): Promise<Output & {
    txid?: string;
}>;
export declare function executeBig<Output>(program: Program, provider: AnchorProvider, command: BigInstructionResult<Output>, payer?: PublicKey, finality?: Finality): Promise<Output & {
    txids?: string[];
}>;
export type PriorityLevel = 'NONE' | 'Min' | 'Low' | 'Medium' | 'High' | 'VeryHigh' | 'UnsafeMax';
type PrioFeeEstimateOpts = {
    includeVote?: boolean;
    recommended?: boolean;
};
export declare const getPriorityFeeEstimate: (connection: Connection, transaction: Buffer, priorityLevel?: PriorityLevel, log?: SplUtilsLogger, options?: PrioFeeEstimateOpts) => Promise<{
    priorityFeeEstimate: any;
}>;
export declare const sendAndConfirmWithRetryAndPrioFees: (connection: Connection, instrs: TransactionInstruction[], signers: Signer[], sendOptions: SendOptions, commitment: Commitment, feePayer?: PublicKey, priorityLevel?: PriorityLevel, timeoutMillis?: number, log?: SplUtilsLogger, addressLookupTableAccounts?: AddressLookupTableAccount[], doNotRetryAt?: Map<number, string>, maxPriorityLevel?: PriorityLevel) => Promise<{
    txid: string;
}>;
export declare const sendMultipleInstructionsWithPrioFee: (provider: AnchorProvider, instructionGroups: TransactionInstruction[][], signerGroups: Signer[][], payer?: PublicKey, priorityLevel?: PriorityLevel, finality?: Finality, timeout?: number, idlErrors?: Map<number, string>, log?: SplUtilsLogger) => Promise<string[]>;
export declare const awaitTransactionSignatureConfirmation: (txid: TransactionSignature, timeoutMillis: number, connection: Connection, commitment?: Commitment, queryStatus?: boolean, log?: SplUtilsLogger) => Promise<SignatureStatus | null | void>;
export declare function sendAndConfirmWithRetry(connection: Connection, txn: Buffer, sendOptions: SendOptions, commitment: Commitment, timeoutMillis?: number, log?: SplUtilsLogger, maxSupportedTransactionVersion?: number): Promise<{
    txid: string;
}>;
export declare function stringToTransaction(solanaTransaction: string): Transaction;
export declare function bufferToTransaction(solanaTransaction: Buffer): Transaction;
type Status = {
    totalProgress: number;
    currentBatchProgress: number;
    currentBatchSize: number;
};
export declare function bulkSendTransactions(provider: Provider, txs: Transaction[], onProgress?: (status: Status) => void, triesRemaining?: number): Promise<string[]>;
export declare function bulkSendRawTransactions(connection: Connection, txs: Buffer[], onProgress?: (status: Status) => void, lastValidBlockHeight?: number, skipPreflight?: boolean, log?: SplUtilsLogger): Promise<string[]>;
export {};
//# sourceMappingURL=transaction.d.ts.map